from BaseHTTPServer import HTTPServer
from BaseHTTPServer import BaseHTTPRequestHandler

from efs.common.encryption import encrypt, decrypt, pubkey_encrypt, pubkey_decrypt, export_key
from efs.common.exceptions import EfsError
from efs.common.exceptions import InvalidRequestError
from efs.common.protocol import Request, Response

from efs.server import fs

class EFSCustomHTTPHandler(BaseHTTPRequestHandler):

    def do_POST(self):
        try:
            content_len = int(self.headers.getheader('Content-Length'))
            (request_encrypted, request_key_id) = self.read_encryption_headers()            
            request_body = self.decrypt_request(self.rfile.read(content_len).decode("utf-8"), 
                request_encrypted, request_key_id)
            efs_request = Request.create_from_text(request_body)
            efs_response = self.handle_efs_request(efs_request).to_json()
            (response_encrypted, response_key_id) = self.get_encryption_values(efs_request)
            response_body = self.encrypt_response(efs_response, 
                response_encrypted, response_key_id).encode("utf-8")
            self.send_response(200)
            self.send_header("Content-Type", "text/plain")
            self.send_header("Content-Length", len(response_body))
            self.send_encryption_headers(response_encrypted, response_key_id)
            self.end_headers()
            self.wfile.write(response_body)
        except EfsError as e:
            self.send_response(400)
            self.send_header("Content-Type", "text/plain")
            self.send_header("Content-Length", 0)
            self.end_headers()

    def decrypt_request(self, request, encrypted, key_id):
        if encrypted == "false":
            return request
        try:
            key = self.server.key_service.get_key(key_id)
            key_type = self.parse_key_id(key_id)
        except:
            raise InvalidRequestError
        if key_type == "server" or key_type=="user":
            return pubkey_decrypt(request, key)
        elif key_type == "session":
            return decrypt(request, key)
        else:
            raise InvalidRequestError

    def encrypt_response(self, response, encrypted, key_id):
        if encrypted == "false":
            return response
        try:
            key = self.server.key_service.get_key(key_id)
            key_type = self.parse_key_id(key_id)
        except:
            raise InvalidRequestError
        if key_type=="user":
            return pubkey_encrypt(response, key)
        elif key_type == "session":
            return encrypt(response, key)
        else:
            raise InvalidRequestError

    def parse_key_id(self, key_id):
        key_parts = key_id.split(":")
        if len(key_parts) > 0:
            return key_parts[0]
        else:
            return key_id

    def read_encryption_headers(self):
        encrypted = self.headers.get("EFS-Encrypted", "false")
        key_id = self.headers.get("EFS-KeyId", "")
        return (encrypted, key_id)

    def get_encryption_values(self, efs_request):
        try:
            user = efs_request.user
            command = efs_request.command.command
        except:
            return ("false", "")
        if command == "get_server_public_key":
            encrypted =  "false"
            key_id = ""
        elif command == "register" or command=="login":
            encrypted = "true"
            key_id = "user:" + efs_request.command.args["user"]
        else:
            encrypted = "true"
            key_id = "session:" + user
        return (encrypted, key_id)

    def send_encryption_headers(self, encrypted, key_id):
        self.send_header("EFS-Encrypted", encrypted)
        self.send_header("EFS-KeyId", key_id)

    def handle_efs_request(self, efs_request):
        #authorize request
        if self.server.authorization_service.authorize(efs_request):
            try:
                command = efs_request.command.command
                handler = getattr(self, "handle_" + command)
            except EfsError as e:
                efs_response = Response(code="InvalidCommandError")
            try:
                efs_response = handler(efs_request)
            except EfsError as e:
                efs_response = Response(code=e.message)

            #update timestamp
            self.server.login_service.update_timestamp(efs_request.user, efs_request.timestamp)
            return efs_response
        
        else:
            return Response(code="UnauthorizedRequestError")

    #Command handlers
    
    def create_error_response(self, e):
        return Response(code=e.__class__.__name__)

    #File system
    
    def handle_mknode(self, efs_request):
        try:
            parent_id = efs_request.command.args["parent_id"]
            perms = efs_request.command.args["perms"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            node_id = fs.mknode(efs_request.user, parent_id, perms)
            return Response(code="Success", data={"node_id":node_id})
        except EfsError as e:
            return self.create_error_response(e)

    def handle_rmnode(self, efs_request):
        try:
            node_id = efs_request.command.args["node_id"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            fs.rmnode(efs_request.user, node_id)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)            

    def handle_read(self, efs_request):
        try:
            node_id = efs_request.command.args["node_id"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            contents = fs.read_node(efs_request.user, node_id)
            return Response(code="Success", data={"contents":contents})
        except EfsError as e:
            return self.create_error_response(e)

    def handle_write(self, efs_request):
        try:
            node_id = efs_request.command.args["node_id"]
            contents = efs_request.command.args["contents"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            fs.write_node(efs_request.user, node_id, contents)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)

    def handle_chmod(self, efs_request):
        try:
            node_id = efs_request.command.args["node_id"]
            perms = efs_request.command.args["perms"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            fs.chmod(efs_request.user, node_id, perms)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)        

    def handle_chgroup(self, efs_request):
        try:
            node_id = efs_request.command.args["node_id"]
            group = efs_request.command.args["group"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            fs.chgrp(efs_request.user, node_id, group)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)

    def handle_mv(self, efs_request):
        try:
            node_id = efs_request.command.args["node_id"]
            new_parent_id = efs_request.command.args["new_parent_id"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            fs.mv_node(efs_request.user, node_id, new_parent_id)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)        

    def handle_cp(self, efs_request):
        try:
            node_id = efs_request.command.args["node_id"]
            copy_parent_id = efs_request.command.args["copy_parent_id"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            new_node_id = fs.cp_node(efs_request.user, node_id, copy_parent_id)
            return Response(code="Success", data={"new_node_id":new_node_id})
        except EfsError as e:
            return self.create_error_response(e)        

    #Login
    
    def handle_register(self, efs_request):
        try:
            user = efs_request.command.args["user"]
            password = efs_request.command.args["password"]
            public_key = efs_request.data["public_key"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            #needs to fix this (what is homedir)
            homedir = self.server.login_service.register(user, password, public_key)
            data = {"homedir": homedir}
            return Response(code="Success", data=data)
        except EfsError as e:
            return self.create_error_response(e)

    def handle_login(self, efs_request):
        try:
            user = efs_request.command.args["user"]
            password = efs_request.command.args["password"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            session = self.server.login_service.login(user, password)
            data = {"token":session['token'], "session_key":session['session_key']}
            return Response(code="Success", data=data)
        except EfsError as e:
            return self.create_error_response(e)

    def handle_logout(self, efs_request):
        try:
            user = efs_request.command.args["user"]
        except:
            return Response(code="InvalidArgumentError")
        self.server.login_service.logout(user)
        return Response(code="Success")

    def handle_chpasswd(self, efs_request):
        try:
            user = efs_request.command.args["user"]
            new_password = efs_request.command.args["new_password"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            self.server.login_service.change_password(user, new_password)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)

    def handle_mkgroup(self, efs_request):
        try:
            group = efs_request.command.args["group"]
            owner = efs_request.command.args["owner"]
            group_public_key = efs_request.command.args["group_public_key"]
            owner_private_key = efs_request.command.args["owner_private_key"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            self.server.login_service.make_group(group, owner, group_public_key, owner_private_key)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)

    def handle_rmgroup(self, efs_request):
        try:
            group = efs_request.command.args["group"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            self.server.login_service.rm_group(group)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)

    def handle_get_groups(self, efs_request):
        try:
            user = efs_request.command.args["user"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            groups = self.server.login_service.get_groups(user)
            return Response(code="Success", data = {"groups":groups})
        except EfsError as e:
            return self.create_error_response(e)

    def handle_adduser(self, efs_request):
        try:
            user = efs_request.command.args["user"]
            group = efs_request.command.args["group"]
            member_private_key = efs_request.data["member_private_key"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            self.server.login_service.add_user(user, group, member_private_key)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)

    def handle_rmuser(self, efs_request):
        try:
            user = efs_request.command.args["user"]
            group = efs_request.command.args["group"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            self.server.login_service.rm_user(user, group)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)

    #Key
    
    def handle_get_server_public_key(self,efs_request):
        key = self.server.key_service.get_server_public_key().publickey()
        data = {"key":export_key(key)}
        return Response(code="Success", data=data)

    def handle_get_user_public_key(self, efs_request):
        try:
            user = efs_request.command.args["user"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            key = self.server.key_service.get_user_public_key(user)
            data = {"key":export_key(key)}
            return Response(code="Success", data=data)
        except EfsError as e:
            return self.create_error_response(e)

    def handle_get_group_public_key(self, efs_request):
        try:
            group = efs_request.command.args["group"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            key = self.server.key_service.get_group_public_key(group)
            data = {"key":export_key(key)}
            return Response(code="Success", data=data)
        except EfsError as e:
            return self.create_error_response(e)

    def handle_get_group_private_key(self, efs_request):
        try:
            user = efs_request.command.args["user"]
            group = efs_request.command.args["group"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            key = self.server.key_service.get_group_private_key(group, user)
            data = {"key":key}
            return Response(code="Success", data=data)
        except EfsError as e:
            return self.create_error_response(e)

    #Struct
    
    def handle_get_user_struct(self, efs_request):
        try:
            user = efs_request.command.args["user"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            struct = self.server.struct_service.get_user_struct(user)
            data = {"struct":struct}
            return Response(code="Success", data=data)
        except EfsError as e:
            return self.create_error_response(e)

    def handle_update_user_struct(self, efs_request):
        try:
            user = efs_request.command.args["user"]
            struct = efs_request.data["struct"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            self.server.struct_service.update_user_struct(user, struct)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)

    def handle_get_group_struct(self, efs_request):
        try:
            group = efs_request.command.args["group"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            struct = self.server.struct_service.get_group_struct(group)
            data = {"struct":struct}
            return Response(code="Success", data=data)
        except EfsError as e:
            return self.create_error_response(e)

    def handle_update_group_struct(self, efs_request):
        try:
            group = efs_request.command.args["group"]
            struct = efs_request.data["struct"]
        except:
            return Response(code="InvalidArgumentError")
        try:
            self.server.struct_service.update_group_struct(group, struct)
            return Response(code="Success")
        except EfsError as e:
            return self.create_error_response(e)

class EFSServer(object):
    def __init__(self, port, login_service, authorization_service, key_service, struct_service):
        self.httpd = HTTPServer(('', port), EFSCustomHTTPHandler)
        self.httpd.login_service = login_service
        self.httpd.authorization_service = authorization_service
        self.httpd.key_service = key_service
        self.httpd.struct_service = struct_service

    def start(self):
        try:
            self.httpd.serve_forever()
        except KeyboardInterrupt:
            self.httpd.shutdown()

    def stop(self):
        self.httpd.shutdown()
