import unittest
from pymongo import MongoClient

from efs.server.fs import init
from efs.common.encryption import generate_pubkey
from efs.server.login import LoginService

class LoginTests(unittest.TestCase):

    def setUp(self):
        self.mongo_client = MongoClient("127.0.0.1", 27017)
        self.users = self.mongo_client.login.users
        self.groups = self.mongo_client.login.groups
        self.sessions = self.mongo_client.login.sessions
        self.users.remove({})
        self.groups.remove({})
        self.sessions.remove({})
        init(self.mongo_client)
        self.login_service = LoginService(self.mongo_client)

    def setUpUser(self):
        private_key = generate_pubkey()
        public_key = private_key.publickey()
        text_public_key = public_key.exportKey()
        self.login_service.register("user1", "12345", text_public_key)

    def test_register(self):
        private_key = generate_pubkey()
        public_key = private_key.publickey()
        text_public_key = public_key.exportKey()
        self.login_service.register("user1", "12345", text_public_key)
        self.assertTrue(self.users.find({"user":"user1"}).count() == 1)

    def test_login_and_logout(self):
        self.setUpUser()
        self.login_service.login("user1", "12345")
        self.assertTrue(self.sessions.find({"user":"user1"}).count() == 1)
        self.login_service.logout("user1")
        self.assertTrue(self.sessions.find({"user":"user1"}).count() == 0)

    def test_change_password(self):
        self.setUpUser()
        self.login_service.change_password("user1", "123456")
        self.login_service.login("user1", "123456")
        self.assertTrue(self.sessions.find({"user":"user1"}).count() == 1)

    def test_make_group(self):
        self.login_service.make_group("mongo", "user1", "xxx", "yyy")
        self.assertTrue(self.groups.find({"group":"mongo"}).count() == 1)

    def test_rm_group(self):
        self.login_service.make_group("mongo", "user1", "xxx", "yyy")
        self.login_service.rm_group("mongo")
        self.assertTrue(self.groups.find({"group":"mongo"}).count() == 0)

    def test_add_user(self):
        self.setUpUser()
        self.login_service.make_group("mongo", "user1", "xxx", "yyy")
        self.login_service.add_user("user1", "mongo", "zzz")
        group_record = self.groups.find({"group":"mongo"})
        self.assertTrue(group_record.count() == 1)
        self.assertTrue("user1" in group_record[0]["members"])
        self.assertEquals("zzz", group_record[0]["members"]["user1"])

    def test_rm_user(self):
        self.setUpUser()
        self.login_service.make_group("mongo", "user1", "xxx", "yyy")
        self.login_service.add_user("user1", "mongo", "zzz")
        self.login_service.rm_user("user1", "mongo")
        self.assertTrue("user1" not in self.groups.find({"group":"mongo"})[0]["members"])

    def tearDown(self):
        self.mongo_client = MongoClient("127.0.0.1", 27017)
        self.users = self.mongo_client.login.users
        self.groups = self.mongo_client.login.groups
        self.sessions = self.mongo_client.login.sessions
        self.users.remove({})
        self.groups.remove({})
        self.sessions.remove({})

if __name__ == '__main__':
    unittest.main()
